#include <jni.h>
#include <string>
#include <android/log.h>
#include <android/window.h>
#include <android/native_window_jni.h>
#include <android/input.h>
#include <android/native_activity.h>
#include "android_native_app_glue.h"
#include "ivr4pengine.h"
#include <EGL/egl.h>
#include <EGL/eglext.h>
#include <unistd.h>
#include <GLES2/gl2.h>

#define LOG_TAG "VR4P"

#define ALOGE(...) __android_log_print( ANDROID_LOG_ERROR, LOG_TAG, __VA_ARGS__ )
#if DEBUG
#define ALOGV(...) __android_log_print( ANDROID_LOG_VERBOSE, LOG_TAG, __VA_ARGS__ )
#else
#define ALOGV(...)
#endif
#include <vector>

using namespace std;

ANativeWindow * g_pNativeWindow = NULL;
EGLDisplay      g_mEglDisplay = NULL;
EGLSurface      g_mEglSurface = EGL_NO_SURFACE;
EGLContext      g_mEglContext = EGL_NO_CONTEXT;

int32_t          g_iSleepTime = 50;

volatile int64_t   m_i64ThisFrameIdx = 1;
volatile int64_t   m_i64ResumeFrameIdx = 0;
volatile int64_t   m_i64PauseFrameIdx = 0;
volatile int64_t   m_i64GameEndDVIdx = 0xfffffff;


double GetTimeInSeconds()
{
    struct timespec now;
    clock_gettime( CLOCK_MONOTONIC, &now );
    return ( now.tv_sec * 1e9 + now.tv_nsec ) * 0.000000001;
}

void app_handle_cmd( struct android_app * app, int32_t cmd )
{
    ALOGE("app_handle_cmd=%d",cmd);

    switch ( cmd )
    {
        case APP_CMD_START:
            break;
        case APP_CMD_RESUME:
        {
            if(g_mEglDisplay == NULL)
                m_i64ResumeFrameIdx = max(m_i64ThisFrameIdx + 1, m_i64PauseFrameIdx + 1);
            break;
        }
        case APP_CMD_PAUSE:
        {
            m_i64PauseFrameIdx = max(m_i64ThisFrameIdx + 4, m_i64ResumeFrameIdx + 1);
            break;
        }
        case APP_CMD_STOP:
            break;
        case APP_CMD_DESTROY:
        {
            g_pNativeWindow = NULL;
            break;
        }
        case APP_CMD_INIT_WINDOW:
        {
            g_pNativeWindow = app->window;
            if(g_mEglDisplay == NULL)
                m_i64ResumeFrameIdx = max(m_i64ThisFrameIdx + 1, m_i64PauseFrameIdx + 1);
            break;
        }
        case APP_CMD_TERM_WINDOW:
        {
            g_pNativeWindow = NULL;
            break;
        }
        default:break;
    }
}

int32_t app_handle_input( struct android_app * app, AInputEvent * event )
{
    if(event != NULL)
    {
        int32_t iCode = AKeyEvent_getKeyCode(event);
        int32_t itype = AInputEvent_getType( event );
        int32_t iaction = AKeyEvent_getAction( event );
        ALOGE("app_handle_input iCode=%d itype=%d iaction=%d",iCode,itype,iaction);
    }
    return 0;
}

#include "Context.h"

void  HandleVrModeChanges()
{
    if(g_pNativeWindow != NULL && m_i64ThisFrameIdx == m_i64ResumeFrameIdx)
    {
        if(g_mEglDisplay == NULL)
        {
            if(V4_IsFrontBuffer())
            {
                if(!InitContext(true,true))
                {
                    UninitContext(true);
                    if(!InitContext(true,false))
                    {
                        UninitContext(true);
                        V4_UseFrontBuffer(false);
                        InitContext(false,false);
                    }
                }
            }
            else
                InitContext(false,false);
        }
    }
    else if(m_i64ThisFrameIdx == m_i64PauseFrameIdx)
    {
        if(g_mEglDisplay != NULL)
            UninitContext(V4_IsFrontBuffer());
    }
}

void  MyVRSleepTime(uint32_t usTime)
{
    struct timespec request;
    if (clock_gettime(CLOCK_REALTIME, &request) == -1)
    {
        usleep((useconds_t) usTime);
        return;
    }
    uint64_t next_vsync = request.tv_nsec + ((uint64_t)usTime) * 1000LL;
    request.tv_sec  += next_vsync / 1000000000;
    request.tv_nsec  = next_vsync % 1000000000;
    clock_nanosleep(CLOCK_REALTIME, TIMER_ABSTIME, &request, NULL);
}


pthread_t           g_android_Render = (pthread_t)NULL;
volatile bool	    g_bGameEndDV = false;
volatile bool	    g_bVRSensorThread = true;

bool                g_bSleepWaitingState = false;
double              g_dSleepWaitingTime = GetTimeInSeconds();
double              g_dTimeForLooping = GetTimeInSeconds();

bool MyIfNeedRealTimeTask()
{
    if(!g_bSleepWaitingState)
        return false;
    double  fFramwWaitTime = double(V4_CfgGetFrameTimeNaS()) / 1000000000.0;
    double  fCompareTime = double(g_iSleepTime) / 2000000.0;
    double thisTime = GetTimeInSeconds();
    return (thisTime - g_dTimeForLooping >= fFramwWaitTime - fCompareTime) && thisTime - g_dSleepWaitingTime >= double(g_iSleepTime) / 1000000.0;
}

void * android_Render(void *)
{
    g_dTimeForLooping = GetTimeInSeconds();
    double  fCompareTime = double(g_iSleepTime) / 2000000.0;

    g_bVRSensorThread = false;

    V4_RegNeedRealTimeTask(MyIfNeedRealTimeTask);

    while (!g_bGameEndDV)
    {
        double fFramwWaitTime = double(V4_CfgGetFrameTimeNaS()) / 1000000000.0;
        if(!V4_IsFrontBuffer() || !V4_CfgGetGlassMode())
        {
            HandleVrModeChanges();
            double thisTime = GetTimeInSeconds();

            if(thisTime - g_dTimeForLooping < fFramwWaitTime - fCompareTime)
            {
                g_dSleepWaitingTime = GetTimeInSeconds();
                g_bSleepWaitingState = true;
                MyVRSleepTime((uint32_t) g_iSleepTime);
                g_bSleepWaitingState = false;
            }
            else
            {
                m_i64ThisFrameIdx++;
                while(thisTime - g_dTimeForLooping >= fFramwWaitTime - fCompareTime * 2.0)
                    g_dTimeForLooping += fFramwWaitTime;

                if(g_mEglDisplay != NULL)
                {
                    if(m_i64PauseFrameIdx <= m_i64ThisFrameIdx + 4 && m_i64PauseFrameIdx >= m_i64ThisFrameIdx)
                    {
                        eglSwapBuffers(g_mEglDisplay,g_mEglSurface);
                        glBindFramebuffer(GL_FRAMEBUFFER, 0);
                        glClearColor(0.0f, 0.0f, 0.0f, 0.0f);
                        glClearStencil(0);
                        glClear(GL_COLOR_BUFFER_BIT|GL_STENCIL_BUFFER_BIT);
                    }
                    else
                    {
                        eglSwapBuffers(g_mEglDisplay,g_mEglSurface);
                        glBindFramebuffer(GL_FRAMEBUFFER, 0);
                        V4_PreRender();

                        if(V4_CfgGetGlassMode())
                        {
                            V4_RenderScene(g_iRenderScene_VRLeft);
                            V4_RenderScene(g_iRenderScene_VRRight);
                        }
                        else
                        {
                            V4_RenderScene(g_iRenderScene_NoGlass);
                        }
                        V4_AftRender();
                    }
                }
            }
        }
        else
        {
            HandleVrModeChanges();

            //  一个周期中，一旦返回true，在调用 V4_PreRender …… V4_RenderScene …… V4_AftRender 前，不能再调用 V4_NeedRenderVR ；
            if(!V4_NeedRenderVR())
            {
                MyVRSleepTime((uint32_t) g_iSleepTime);
            }
            else
            {
                m_i64ThisFrameIdx++;
                if(g_mEglDisplay != NULL)
                {
                    if(m_i64PauseFrameIdx <= m_i64ThisFrameIdx + 4 && m_i64PauseFrameIdx >= m_i64ThisFrameIdx)
                    {
                        eglSwapBuffers(g_mEglDisplay,g_mEglSurface);
                        glBindFramebuffer(GL_FRAMEBUFFER, 0);
                        glClearColor(0.0f, 0.0f, 0.0f, 0.0f);
                        glClearStencil(0);
                        glClear(GL_COLOR_BUFFER_BIT|GL_STENCIL_BUFFER_BIT);
                    }
                    else
                    {
                        eglSwapBuffers(g_mEglDisplay,g_mEglSurface);
                        glBindFramebuffer(GL_FRAMEBUFFER, 0);
                        V4_PreRender();
                        V4_RenderScene(g_iRenderScene_VRLeft);
                        V4_RenderScene(g_iRenderScene_VRRight);
                        V4_AftRender();
                    }
                }
            }
        }
    }

    m_i64PauseFrameIdx = m_i64ThisFrameIdx;
    HandleVrModeChanges();

    g_bVRSensorThread = true;
    pthread_exit((void *)0);
}


JavaVM *	g_Vm = NULL;
JNIEnv *	g_Env = NULL;

#include <sys/prctl.h>

void android_main( struct android_app * app )
{
    ANativeActivity_setWindowFlags( app->activity, AWINDOW_FLAG_KEEP_SCREEN_ON, 0 );
    g_Vm = app->activity->vm;
    (*g_Vm).AttachCurrentThread(&g_Env, NULL );

    prctl( PR_SET_NAME, (long)"VR4PNATIVE::Main", 0, 0, 0 );

    app->userData = NULL;
    app->onAppCmd = app_handle_cmd;
    app->onInputEvent = app_handle_input;

    m_i64ThisFrameIdx = 1;
    m_i64ResumeFrameIdx = 0;
    m_i64PauseFrameIdx = 0;
    m_i64GameEndDVIdx = 0xfffffff;

    pthread_create(&g_android_Render, NULL, android_Render, NULL);

    while ( app->destroyRequested == 0 )
    {
        for (;;)
        {
            int events;
            struct android_poll_source *source;
            if (ALooper_pollAll(-1, NULL, &events, (void **) &source) < 0)
                break;
            if (source != NULL)
                source->process(app, source);
        }
    }

    g_bGameEndDV = true;
    while(!g_bVRSensorThread)
        pthread_join(g_android_Render,NULL);

    (*g_Vm).DetachCurrentThread();
}





